import datetime
from typing import Optional


DATE_TIME_FORMAT = '%d-%m-%y %H:%M:%S.%f%z'


def serialize_datetime(passed_datetime: Optional[datetime.datetime]) -> Optional[str]:
    """
    Serialize a datetime instance, which may be None, to a string representation
    If datetime is None, None is returned
    """

    if isinstance(passed_datetime, datetime.datetime):
        return passed_datetime.strftime(DATE_TIME_FORMAT)
    return None


def deserialize_datetime(str_datetime: Optional[str]) -> Optional[datetime.datetime]:
    """
    Deserializes a string of a date to a datetime format
    if str_datetime is 'falsy', a value of None is returned
    """
    if not str_datetime:
        return None

    return datetime.datetime.strptime(str_datetime, DATE_TIME_FORMAT)


from django.test import RequestFactory
from django.contrib.sessions.middleware import SessionMiddleware
from rest_framework.response import Response
from django.core.cache import cache
from unittest.mock import patch, AsyncMock
import asyncio
import pytest
from bgs_chat_app.views import init_conversation, send_message, get_messages
from ..common.fern import Fern
from ..common.api_calls import POLLING_INTERVAL


@pytest.mark.django_db
def test_init_generates_token_and_id():
    """
    Checks that the views init_conversation generates a (mocked) token and
    conversation id;
    Checks that a conversation key is generated and stored in session against
    a hashed conversation key
    """
    factory = RequestFactory()
    request = factory.get('/chat')

    middleware = SessionMiddleware(lambda req: None)
    middleware.process_request(request)
    request.session.save()
    with patch('bgs_chat_app.common.api_calls.DirectLineAPI.get_token_and_conversation_id', return_value = {
            'conversation_id': 'test_conversation_id',
            'token': 'test_token',
        }
    ):
        init_conversation(request)

    # this is the hashed conversation key
    assert 'conversation_key' in request.session.keys()
    decrypted_key = Fern.decrypt(request.session['conversation_key'])
    bot_info = cache.get(f'chat_bot_{decrypted_key}')
    assert bot_info['conversation_id'] == 'test_conversation_id'
    assert bot_info['token'] == 'test_token'


@pytest.mark.django_db
def test_send_message():
    """
    Test that bot.send_message a message returns ({'status': 'message sent'})
    """
    factory = RequestFactory()
    request = factory.post('/chat', data={"message": "This is a message"}, content_type="application/json")

    middleware = SessionMiddleware(lambda req: None)
    middleware.process_request(request)
    request.session.save()
    with patch('bgs_chat_app.common.api_calls.DirectLineAPI.get_token_and_conversation_id', return_value={
            'conversation_id': 'test_conversation_id',
            'token': 'test_token',
        }), \
        patch('bgs_chat_app.common.api_calls.DirectLineAPI.send_message', return_value={
            'status': 'message sent'
            }):
        init_conversation(request)
    res = send_message(request)
    assert res.status_code == 200
    assert res.data['status'] == "message sent"


@pytest.mark.django_db
@pytest.mark.asyncio
def test_get_message():
    """
    Test that get_response returns a response from the bot
    """
    factory = RequestFactory()
    request = factory.get('/chat')
    middleware = SessionMiddleware(lambda req: None)
    middleware.process_request(request)
    request.session.save()
    with patch('bgs_chat_app.common.api_calls.DirectLineAPI.start_conversation') as mock_start, \
        patch('bgs_chat_app.common.api_calls.DirectLineAPI.send_message') as mock_send, \
        patch('bgs_chat_app.common.api_calls.DirectLineAPI.get_response', new_callable=AsyncMock) as mock_get_response:

        mock_start.return_value = {"conversation_id": 'fake-id'}
        mock_send.return_value = None
        mock_get_response.side_effect = [
            asyncio.sleep(POLLING_INTERVAL),
            "Hello, how can I help you today?"
        ]

    response = get_messages(request)

    assert response.status_code == 400
    assert response.data['error'] == 'No active conversation found'

# def test_send_message_no_credentials
"""
Test that a message sent with no credentials returns the expected error message
"""


# @pytest.mark.django_db
# def test_send_message_throttled(settings):
#     """
#     Test that bot.send_message throttling is working correctly
#     """
#     # assert 0
#     settings.REST_FRAMEWORK["DEFAULT_THROTTLE_RATES"]["anon"] = "1/hour"
#     factory = RequestFactory()
#     request = factory.post('/chat', data={"message": "This is a message"}, content_type="application/json")

#     middleware = SessionMiddleware(lambda req: None)
#     middleware.process_request(request)
#     request.session.save()
#     with patch('bgs_chat_app.common.api_calls.DirectLineAPI.get_token_and_conversation_id', return_value={
#             'conversation_id': 'test_conversation_id',
#             'token': 'test_token',
#         }), \
#         patch('bgs_chat_app.common.api_calls.DirectLineAPI.send_message', return_value={
#             'status': 'message sent'
#             }):
#         init_conversation(request)
#     # # Check the first message sends correctly
#     res = send_message(request)
#     assert res.status_code == 200
#     assert res.data['status'] == "message sent"

#     new_request = factory.post('/chat', data={"message": "This is another message"}, content_type="application/json")
#     middleware.process_request(new_request)
#     new_request.session.save()
#     with patch('bgs_chat_app.common.api_calls.DirectLineAPI.get_token_and_conversation_id', return_value={
#             'conversation_id': 'test_conversation_id',
#             'token': 'test_token',
#         }), \
#         patch('bgs_chat_app.common.api_calls.DirectLineAPI.send_message', return_value={
#             'status': 'message sent'
#             }):
#         init_conversation(new_request)
#     res_2 = send_message(new_request)

#     assert res_2.status_code == 429

#     settings.REST_FRAMEWORK["DEFAULT_THROTTLE_RATES"]["anon"] = "100/hr"


from bgs_chat.common.api_calls import DirectLineAPI
import json
import datetime
from rest_framework import status
from rest_framework.decorators import api_view, throttle_classes
from rest_framework.throttling import AnonRateThrottle
from rest_framework.response import Response
import uuid
from django.core.cache import cache
from bgs_chat.common.serializers import serialize_datetime, deserialize_datetime
from bgs_chat.common.fern import Fern


def index(request):
    init_conversation()
    return {'message': 'Conversation started'}

@api_view(['POST'])
def init_conversation(request):
    """
    Initialise the chatbot, generate a conversation_id and individual token.
    Generate a conversation_key to reference the conversation_id and token.
    Cache the conversation_id and token so it is isolated from the front end.
    """
    print('*'*30)
    print('VIEWS/INIT_CONVERSATION')
    print('*'*30)
    bot = DirectLineAPI()
    token_and_id = bot.get_token_and_conversation_id()
    conversation_key = str(uuid.uuid4())
    # cache the token_and_id against a uuid conversation_key

    cache.set(f'chat_bot_{conversation_key}', token_and_id, timeout=3600)
    # hash the conversation key before saving to session
    hashed = Fern.encrypt(conversation_key)
    request.session['conversation_key'] = hashed
    bot.conversation_id = token_and_id['conversation_id']
    bot._token = token_and_id['token']
    bot.start_conversation()
    request.session.modified = True
    return Response({'text': 'Token successfully generated'}, status=200)


def get_bot_info(hashed_conversation_key):
    conversation_key = Fern.decrypt(hashed_conversation_key)
    if conversation_key is None:
        return Response({"error": "Error decrypting conversation key, please contact support"})
    return cache.get(f'chat_bot_{conversation_key}')


# We may want to reset this as an api_view...
# @api_view(['POST'])
def refresh_token(request):
    hashed_conversation_key = request.session.get('conversation_key')
    if not hashed_conversation_key:
        init_conversation(request)

    bot_info = get_bot_info(hashed_conversation_key)
    if not bot_info:
        request.session['conversation_key'] = None

        return Response({"error": "Bot info not found, conversation key deleted. Please start a new conversation"}, status=400)

    bot = DirectLineAPI(
        token=bot_info['token'],
        conversation_id=bot_info['conversation_id']
    )

    res = bot.refresh_token()
    if res['status_code'] == 200:
        conversation_key = Fern.decrypt(hashed_conversation_key)
        cache.set(f'chat_bot_{conversation_key}', res['token_and_id'], timeout=3600)
        return Response({"error": "Token refreshed, your new token is valid for one hour"}, status=200)
    else:
        return Response({"error": f"Could not refresh token {res['status_code']} - {res['content']}"}, status=400)


@api_view(['POST'])
@throttle_classes([AnonRateThrottle])
def send_message(request):
    print('*'*30)
    print('BGS_CHAT_APP/VIEWS.PY SEND_MESSAGE')
    print('*'*30)
    hashed_conversation_key = request.session.get('conversation_key')

    if not hashed_conversation_key:
        init_conversation(request)
        hashed_conversation_key = request.session.get('conversation_key')

    bot_info = get_bot_info(hashed_conversation_key)

    if not bot_info:
        request.session['conversation_key'] = None
        init_conversation(request)
        hashed_conversation_key = request.session.get('conversation_key')
        bot_info = get_bot_info(hashed_conversation_key)
        bot = DirectLineAPI(
            token=bot_info['token'],
            conversation_id=bot_info['conversation_id'],
        )
        bot.start_conversation()

    message = json.loads(request.body.decode('utf-8')).get("message")

    if not message:
        return Response({"error": "Missing message"})

    last_message_sent_at = request.session.get('last_message_sent_timestamp', datetime.datetime.now(datetime.timezone.utc)) or datetime.datetime.now(datetime.timezone.utc)

    bot = DirectLineAPI(
        token=bot_info['token'],
        conversation_id=bot_info['conversation_id'],
        last_message_sent_timestamp=last_message_sent_at,
    )

    bot.send_message(message)
    request.session['last_message_sent_timestamp'] = serialize_datetime(last_message_sent_at)
    request.session.modified = True

    return Response({'status': 'message sent'})


@api_view(['GET'])
def get_messages(request):
    hashed_conversation_key = request.session.get('conversation_key')

    if not hashed_conversation_key:
        return Response({"error": "No active conversation found"}, status=400)

    bot_info = get_bot_info(hashed_conversation_key)

    if not bot_info:
        # NOTE: If there's no bot info for get_response, we should prompt the
        # user to re-enter their question/raise an error
        return Response({"error": "Bot info not found"}, status=400)

    last_message_sent_at = deserialize_datetime(request.session['last_message_sent_timestamp'])
    watermark = request.session.get('watermark', None)

    bot = DirectLineAPI(
        token=bot_info['token'],
        conversation_id=bot_info['conversation_id'],
        last_message_sent_timestamp=last_message_sent_at,
        watermark=watermark
    )
    response_info = bot.get_response()
    request.session['watermark'] = bot.watermark

    request.session.modified = True
    # import pdb; pdb.set_trace()
    return Response({"text": response_info['text']}, status=status.HTTP_200_OK)